/* $Id: sflow.c,v 1.11 2003/01/04 11:27:53 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "interp.h"
#include "shared.h"
#include "flow.h"
#define SFLOW_IMP
#include "sflow.h"
#include "util.h"

flow_class script_flow_class = {"script", sf_open, sf_close, sf_getc, sf_ungetc, 
							   sf_putc, sf_puts,
							   sf_getpos, sf_setpos, sf_rewind,
							   sf_flowctl
							 };


typedef struct script_flow_data
{
	unsigned char writeable : 1;
	interpreter *filter;
	procedure *getc_proc, 
		*putc_proc, 
		*puts_proc, 
		*ioctl_proc, 
		*ungetc_proc, 
		*getpos_proc, 
		*setpos_proc,
		*rewind_proc;
	flow_class *nested_cls;
	flow_handle_t nested_data;
} script_flow_data;

flow_handle_t sf_open(char *name, char *options, int writeable)
{
	char *script_param = strchr(options, '?');
	script_flow_data *sfd = ALLOC(script_flow_data);
	dstring *tmp;

	if(script_param)
		*script_param = '\0';
	sfd->writeable = writeable;
	sfd->filter = find_exported(SUBINT_CLASS, options);
	if(!sfd->filter)
	{
		FREE(sfd);
		recover_error("can't find exported interpreter '%s'", options);
	}
	tmp = eval_procedurev(sfd->filter, "openflow", 
						 ds_fromptr(sfd), 
						 ds_create(script_param), ds_create(writeable ? "true" : ""), 
						 ds_create(name),
						 NULL);
	if(tmp)
	{
		ds_unfix(tmp);
		if(!ds_isempty(tmp))
			recover_error("can't open script flow %s: %s", name, DS_BODY(tmp));
	}
	sfd->getc_proc = qfind_proc(sfd->filter, "getc");
	sfd->putc_proc = qfind_proc(sfd->filter, "putc");
	sfd->puts_proc = qfind_proc(sfd->filter, "puts");
	sfd->ioctl_proc =  qfind_proc(sfd->filter, "ioctl");
	sfd->ungetc_proc = qfind_proc(sfd->filter, "ungetc");
	sfd->getpos_proc = qfind_proc(sfd->filter, "getpos");
	sfd->setpos_proc = qfind_proc(sfd->filter, "setpos");
	sfd->rewind_proc = qfind_proc(sfd->filter, "rewind");
	return sfd;
}

void sf_close(flow_handle_t fh)
{
	script_flow_data *sfd = fh;
	discard_value(eval_procedurev(sfd->filter, "closeflow", 
						  ds_fromptr(fh), 
						  NULL));
	FREE(sfd);
}

int sf_getc(flow_handle_t fh)
{
	script_flow_data *sfd = fh;
	dstring *tmp;
	int res;
	
	if(!sfd->getc_proc)
		return FLOW_EOF;
	tmp = eval_procedurepv(sfd->filter, sfd->getc_proc, 
						   ds_fromptr(fh), 
						   NULL); 
	if(!tmp)
		return FLOW_EOF;
	ds_unfix(tmp);
	if(!*DS_BODY(tmp))
		return FLOW_EOF;
	res = convert_to_int(DS_BODY(tmp));
	return res >= 0 ? res : FLOW_ERROR;
}

int sf_ungetc(flow_handle_t fh, int ch)
{
	script_flow_data *sfd = fh;
	dstring *tmp;

	if(!sfd->ungetc_proc)
		return FLOW_ERROR;
	tmp = eval_procedurepv(sfd->filter, sfd->ungetc_proc, ds_fromptr(fh), ds_fromint(ch, 10, 0), NULL);
	if(!tmp)
		return 0;
	ds_unfix(tmp);
	return ds_isempty(tmp) ? 0 : FLOW_ERROR;
}


int sf_putc(flow_handle_t fh, int ch)
{
	script_flow_data *sfd = fh;
	dstring *tmp;
	
	if(!sfd->writeable || !sfd->putc_proc)
	{
		errno = EACCES;
		return FLOW_ERROR;
	}
	tmp = eval_procedurepv(sfd->filter, sfd->putc_proc, 
						  ds_fromptr(fh), 
						  ds_fromint(ch, 10, 0), 
						  NULL);
	if(tmp)
	{
		ds_unfix(tmp);
		return ds_isempty(tmp) ? 0 : FLOW_ERROR;
	}
	return 0;
}

int sf_puts(flow_handle_t fh, char *str)
{
	script_flow_data *sfd = fh;
	dstring *tmp;
	if(!str)
	{
		errno = EINVAL;
		return FLOW_ERROR;
	}
	if(!sfd->writeable)
	{
		errno = EACCES;
		return FLOW_ERROR;
	}
	if(!sfd->puts_proc)
	{
		int res;
		for(; *str; str++)
		{
			if(res = sf_putc(fh, *str))
				return res;
		}
		return 0;
	}
	tmp = eval_procedurepv(sfd->filter, sfd->puts_proc, 
						  ds_fromptr(fh), 
						  ds_create(str), NULL);
	if(tmp)
	{
		ds_unfix(tmp);
		return ds_isempty(tmp) ? 0 : FLOW_ERROR;
	}
	return 0;
}
 
flow_pos_t sf_getpos(flow_handle_t fh)
{
	script_flow_data *sfd = fh;
	if(sfd->getpos_proc)
	{
		dstring *tmp = eval_procedurepv(sfd->filter, sfd->getpos_proc, ds_fromptr(fh), NULL);
		if(!tmp)
			return FLOW_ERROR;
		ds_unfix(tmp);
		return convert_to_int(DS_BODY(tmp));
	}
	return FLOW_ERROR;
}

int sf_setpos(flow_handle_t fh, flow_pos_t fp)
{
	script_flow_data *sfd = fh;
	if(sfd->setpos_proc)
	{
		dstring *tmp = eval_procedurepv(sfd->filter, sfd->setpos_proc, ds_fromptr(fh), ds_fromint(fp, 10, 0), NULL);
		if(!tmp)
			return 0;
		ds_unfix(tmp);
		return ds_isempty(tmp) ? 0 : FLOW_ERROR;
	}
	return FLOW_ERROR;
}

int sf_rewind(flow_handle_t fh)
{
	script_flow_data *sfd = fh;
	if(sfd->rewind_proc)
	{
		dstring *tmp = eval_procedurepv(sfd->filter, sfd->rewind_proc, ds_fromptr(fh), NULL);
		if(!tmp)
			return 0;
		ds_unfix(tmp);
		return ds_isempty(tmp) ? 0 : FLOW_ERROR;
	}
	else return sf_setpos(fh, 0);
}

char *sf_flowctl(flow_handle_t fh, char *ctl)
{
	script_flow_data *sfd = fh;
	dstring *tmp;
	if(!sfd->ioctl_proc)
		return NULL;
	tmp = eval_procedurepv(sfd->filter, sfd->ioctl_proc, ds_fromptr(fh), ds_create(ctl), NULL);
	if(tmp)
	{
		ds_unfix(tmp);
		if(!ds_comparestr(tmp, "not-implemented", NULL, NULL))
			return NULL;
		return DS_BODY(tmp);
	}
	return "";
}
